#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/common/common.h>
#include <pcl/common/angles.h>
#include <pcl/common/intersections.h>
#include <pcl/common/geometry.h>
#include <pcl/features/organized_edge_detection.h>
#include <iostream>
#include <vector>
#include <math.h>
#include <string>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/crop_hull.h>
#include <pcl/surface/convex_hull.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/filters/project_inliers.h>

#include <pcl/visualization/cloud_viewer.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/segmentation/extract_clusters.h>

#include <pcl/common/centroid.h>
#include <pcl/common/transforms.h>
#include <pcl/features/moment_of_inertia_estimation.h>
#include <pcl/filters/radius_outlier_removal.h>
#include <pcl/filters/project_inliers.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/surface/concave_hull.h>
#include <pcl/filters/passthrough.h>

class BinSegmentation
{

private:
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr m_source;
    pcl::PointCloud<pcl::PointXYZ>::Ptr m_source_bw, m_edges_hull, m_hull_result;

    // edges
    pcl::PointCloud<pcl::PointXYZ>::Ptr m_occluding_edges, m_occluding_edges_output;
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr m_occluded_edges;
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr m_boundary_edges;

    pcl::PointCloud<pcl::PointXYZ>::Ptr m_cloud_vertices, m_cloud_vertices_scaled;
    pcl::ModelCoefficients::Ptr m_plane;

    std::vector<pcl::ModelCoefficients::Ptr> models;
    std::vector<pcl::ModelCoefficients::Ptr> m_models;
    std::vector<pcl::ModelCoefficients> m_lines;
    std::vector<Eigen::Vector3f> m_bin_corners;

    int m_num_lines;
    double m_sqr_eps; //  maximum allowable distance to the true solution lineWithLineIntersection (need to be taken squared) and
                      // maximum distance between point and line in checkOrthogonality
    float m_safe_distance;

public:
    BinSegmentation() : m_source(new pcl::PointCloud<pcl::PointXYZRGB>),
                        m_source_bw(new pcl::PointCloud<pcl::PointXYZ>),
                        m_edges_hull(new pcl::PointCloud<pcl::PointXYZ>),
                        m_occluding_edges(new pcl::PointCloud<pcl::PointXYZ>),
                        m_occluding_edges_output(new pcl::PointCloud<pcl::PointXYZ>),
                        m_occluded_edges(new pcl::PointCloud<pcl::PointXYZRGBA>),
                        m_boundary_edges(new pcl::PointCloud<pcl::PointXYZRGBA>),
                        m_cloud_vertices(new pcl::PointCloud<pcl::PointXYZ>),
                        m_cloud_vertices_scaled(new pcl::PointCloud<pcl::PointXYZ>),
                        m_hull_result(new pcl::PointCloud<pcl::PointXYZ>),
                        m_plane(new pcl::ModelCoefficients)
    {
        m_sqr_eps = 0.03;
        m_num_lines = 4;
        m_safe_distance = 0;
    }

    void setInputCloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud);
    void setInputBinPoints(std::vector<Eigen::Vector3f> &bin_points) {m_bin_corners = bin_points;}

    void setSafeDistance(float scale) { m_safe_distance = scale; }

    pcl::ModelCoefficients::Ptr getReferencePlane() { return m_plane; };

    pcl::PointCloud<pcl::PointXYZ>::Ptr getEdgesCloud();

    bool compute(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_entropy);

    void visualize();

private:
    bool computeEdges(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud,
                      pcl::PointCloud<pcl::PointXYZ>::Ptr &occluding_edges,
                      pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &occluded_edges,
                      pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &boundary_edges);

    bool ransacLineDetection(pcl::PointCloud<pcl::PointXYZ>::Ptr &occluding_edges,
                             std::vector<pcl::ModelCoefficients> &lines);

    bool compareWithPreviousLines(std::vector<pcl::ModelCoefficients> &lines, pcl::ModelCoefficients &line);

    bool checkLinesOrthogonal(std::vector<pcl::ModelCoefficients> &lines,
                              std::vector<Eigen::Vector4f> &points);

    bool getIntersactions(std::vector<pcl::ModelCoefficients> &lines,
                          pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_vertices);

    pcl::ModelCoefficients::Ptr planeModel(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud);

    bool segmentOccludingEdges(pcl::PointCloud<pcl::PointXYZ>::Ptr &occluding_edges,
                               pcl::ModelCoefficients::Ptr &plane);

    void scaleVertices(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_vertices,
                       pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_vertices_scaled);

    bool addPoints(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_vertices_scaled,
                   pcl::ModelCoefficients::Ptr &plane,
                   float height_value);

    void convexHullCrop(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_bw,
                        pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_vertices,
                        pcl::PointCloud<pcl::PointXYZ>::Ptr &hull_result);

    void updatePlane(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::ModelCoefficients::Ptr &plane);

    void edgesHull(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_edges,
                   pcl::ModelCoefficients::Ptr &plane,
                   pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_hull);
};