#include "ros/ros.h"
#include <sensor_msgs/PointCloud2.h>
#include <pcl_conversions/pcl_conversions.h>
#include <geometry_msgs/Pose.h>
#include <eigen_conversions/eigen_msg.h> //tf::poseEigenToMsg

#include "ros/ros.h"
#include <geometry_msgs/Pose.h>
#include <sensor_msgs/PointCloud2.h>
#include <eigen_conversions/eigen_msg.h>
#include <tf2_ros/transform_listener.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <tf2_ros/static_transform_broadcaster.h>
#include <geometry_msgs/TransformStamped.h>
#include <tf2/LinearMath/Quaternion.h>
#include <std_msgs/Float64MultiArray.h>
#include "pcl_ros/point_cloud.h"
#include <pcl/common/transforms.h>
#include <Eigen/Geometry>
#include "tf2_eigen/tf2_eigen.h"

#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/common/time.h>
#include <pcl/console/print.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/filters/passthrough.h>


#include "vision_msgs/SegCloud.h"
#include "../include/bin_segmentation.h"



class Segment{

    std::string topic_cloud_, topic_srv_;
    float bin_safe_distance_;

    ros::NodeHandle n_;
    ros::ServiceServer get_srv_;

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_;

public:
    Segment(ros::NodeHandle* n);

    void maincall();

    void cloudCallback(const pcl::PointCloud<pcl::PointXYZ>::ConstPtr& msg);

    bool serviceCallback(   vision_msgs::SegCloud::Request &req,
                            vision_msgs::SegCloud::Response &res);
};

Segment::Segment(ros::NodeHandle* n):n_(*n) , cloud_(new pcl::PointCloud<pcl::PointXYZRGB>)
{
    std::cout<<"Constructor started"<<std::endl;
    
    n_.getParam("bin_seg/topic_cloud", topic_cloud_);
    n_.getParam("bin_seg/topic_srv", topic_srv_);
    n_.getParam("bin_seg/bin_safe_distance", bin_safe_distance_);

    get_srv_ = n_.advertiseService(topic_srv_, &Segment::serviceCallback, this);

    std::cout<<"Constructor finished"<<std::endl;
}

bool Segment::serviceCallback(vision_msgs::SegCloud::Request &req, vision_msgs::SegCloud::Response &res){
    
    pcl::console::print_highlight("Getting source cloud...\n");
    sensor_msgs::PointCloud2ConstPtr sharedMsg;
    sensor_msgs::PointCloud2 msg;
    sharedMsg = ros::topic::waitForMessage<sensor_msgs::PointCloud2>(topic_cloud_, n_);
    if (sharedMsg != NULL)
        msg = *sharedMsg;

    pcl::PCLPointCloud2 pcl_pc2;
    pcl_conversions::toPCL(*sharedMsg, pcl_pc2);
    pcl::fromPCLPointCloud2(pcl_pc2, *cloud_);
    pcl::console::print_highlight("done\n");

    // ----------- BIN SEGMENTATION -----------
    BinSegmentation bin;
    bin.setInputCloud(cloud_);
    bin.setSafeDistance(bin_safe_distance_);

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_entropy(new pcl::PointCloud<pcl::PointXYZRGB>);
    if (!bin.compute(cloud_entropy)){
        return -1;
    }
    pcl::ModelCoefficients::Ptr plane = bin.getReferencePlane();
    std::cout << plane->values[0] << " " << plane->values[1] << " " << plane->values[2] << " " << plane->values[3] << std::endl;

    // ----------- CLOUD PRE-PROCESSING -----------
    pcl::PassThrough<pcl::PointXYZRGB> pass;
    pass.setInputCloud(cloud_entropy);
    pass.setFilterFieldName("z");
    pass.setFilterLimits(0.0, fabs(plane->values[3]));
    pass.filter(*cloud_entropy);


    // TRANSFORM TO POINTCLOUD2 TYPE
    sensor_msgs::PointCloud2 cloud_output;
    pcl::toROSMsg(*cloud_entropy, cloud_output);

    // TRANSFROM PLANE TO FLOAT64 VECTOR
    std::vector<double> plane_output = {plane->values[0], plane->values[1], plane->values[2], plane->values[3]};

    res.cloud_seg = cloud_output;
    res.plane = plane_output;

    return true;


}


// Align a rigid object to a scene with clutter and occlusions
int main(int argc, char **argv)
{
    ros::init(argc, argv, "bin_segmentation_node");
    ros::NodeHandle n;

    Segment s(&n);

    ROS_INFO("Service Ready");
    ros::spin();
 
    return (0);
}
