#include "ros/ros.h"
#include <geometry_msgs/PoseArray.h>

#include "vision_msgs/SegCloud.h"
#include "vision_msgs/WrinkledPosesEstimator.h"


class Supervisor{
    ros::NodeHandle n_;
    ros::ServiceClient client_seg_, client_estimator_;
    std::string topic_seg_srv_, topic_est_srv_;

public:
    Supervisor(ros::NodeHandle* n);
    int maincall();
};

Supervisor::Supervisor(ros::NodeHandle* n):n_(*n) 
{
    std::cout<<"Constructor started"<<std::endl;

    n_.getParam("wm_supervisor/topic_seg_srv", topic_seg_srv_);
    n_.getParam("wm_supervisor/topic_est_srv", topic_est_srv_);

    client_seg_ = n_.serviceClient<vision_msgs::SegCloud>(topic_seg_srv_);
    client_estimator_ = n_.serviceClient<vision_msgs::WrinkledPosesEstimator>(topic_est_srv_);

    std::cout<<"Constructor finished"<<std::endl;
}


int Supervisor::maincall(){
    std::cout<<"inside main call"<<std::endl;

    vision_msgs::SegCloud seg_srv;
    if (client_seg_.call(seg_srv))
    {
        ROS_INFO("Called <segment_cloud_bin> service!");

        vision_msgs::WrinkledPosesEstimator estimator_srv;
        estimator_srv.request.cloud_seg = seg_srv.response.cloud_seg;
        estimator_srv.request.plane = seg_srv.response.plane;
        estimator_srv.request.task_id = 2; // WM
        if (client_estimator_.call(estimator_srv))
        {
            ROS_INFO("Called <estimate_wrinkled_poses> service!");
            geometry_msgs::PoseArray output;
            output = estimator_srv.response.poses;
        }
        else
        {
            ROS_ERROR("Failed to call service <estimate_wrinkled_poses>!");
            return 1;
        }
    }
    else
    {
        ROS_ERROR("Failed to call service <segment_cloud_bin>!");
        return 1;
    }

}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "supervisor_wm");
    ros::NodeHandle n;
    Supervisor v(&n);
    
    v.maincall();

    return 0;
}
