#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/filters/project_inliers.h>
#include <pcl/search/kdtree.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/common/centroid.h>
#include <pcl/common/common.h>
#include <pcl/surface/concave_hull.h>
#include <pcl/filters/crop_hull.h>
#include <pcl/surface/convex_hull.h>
#include <pcl/common/intersections.h>
#include <pcl/segmentation/extract_clusters.h>
// Types
typedef pcl::PointNormal PointNT;
typedef pcl::PointXYZI PointT;
typedef pcl::PointCloud<PointNT> PointCloudT;
typedef pcl::PointCloud<PointT> PointCloudIntT;

class WMSegmentation
{
private:
    pcl::ModelCoefficients::Ptr plane_;
    PointCloudT::Ptr scene_, object_;
    PointCloudIntT::Ptr scene_map_;
    PointCloudT::Ptr cloud_hull_;

    float threshold_seg_;

public:
    WMSegmentation() : plane_(new pcl::ModelCoefficients),
                       object_(new PointCloudT),
                       scene_(new PointCloudT),
                       scene_map_(new PointCloudIntT),
                       cloud_hull_(new PointCloudT)
    {
        threshold_seg_ = 0.025;
    }

    bool compute(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_seg);

    void setSceneCloud(PointCloudT::Ptr scene) { scene_ = scene; };

    void setObjectCloud(PointCloudT::Ptr object) { object_ = object; };

    void setSegmentationThreshold(float th) { threshold_seg_ = th; };

    pcl::ModelCoefficients::Ptr getPlane() { return plane_; };

    PointCloudIntT::Ptr getSceneMap() { return scene_map_; };

    void visualize(bool flag_spin = false);

private:
    void removePerperndicularPlane(PointCloudT::Ptr &cloud,
                                   pcl::ModelCoefficients::Ptr &plane,
                                   PointCloudT::Ptr &cloud_filtered);

    void getPointPlaneDistanceCloud(PointCloudT::Ptr &obj,
                                    PointCloudT::Ptr &scene,
                                    PointCloudT::Ptr scene_proj,
                                    PointCloudIntT::Ptr &cloud_out,
                                    pcl::ModelCoefficients::Ptr &plane);

    void computeDiffProjected(PointCloudT::Ptr obj,
                              PointCloudT::Ptr scene,
                              PointCloudIntT::Ptr cloud_output,
                              PointCloudT::Ptr hull,
                              pcl::ModelCoefficients::Ptr &plane);

    void segmentSceneCloud(PointCloudIntT::Ptr cloud, pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_seg, float threshold);
};