#include "../include/spline.h"

std::vector<Eigen::Vector3f> Spline::SplineApproximation(std::vector<Eigen::Vector3f> path)
{

    std::vector<Eigen::Vector3f> result_points;
    int degree = std::min<int>(path.size() - 1, 2);
    // The degree of the interpolating spline needs to be one less than the number of points
    // that are fitted to the spline.
    Eigen::MatrixXf points(3, path.size());
    int row_index = 0;
    for (auto const way_point : path)
    {
        points.col(row_index) << way_point[0], way_point[1], way_point[2];
        row_index++;
    }
    std::cout << "points cols: " << points.cols() << std::endl;
    std::cout << "spline degree: " << degree << std::endl;
    Spline3d spline = Eigen::SplineFitting<Spline3d>::Interpolate(points, degree);

    /*
    float time_ = 0;
    for (int i = 0; i < 20; i++)
    {
        time_ += 1.0 / (20 * 1.0);
        Eigen::Vector3f values = spline(time_);
        std::cout << values.x() << ", " << values.y() << ", " << values.z() << std::endl;
        //result_points.push_back(values);
    }
    */

    Eigen::Array<float, 1, -1> knots = spline.knots();
    Eigen::Spline<float, 3, Eigen::Dynamic> Bspline(knots, points);

    float time_ = 0;
    int num_points_ = 200;
    for (int i = 0; i < num_points_; i++)
    {
        time_ += 1.0 / (num_points_ * 1.0);
        result_points.push_back(Bspline(time_));
    }

    return result_points;
}

pcl::PointCloud<pcl::PointXYZ>::Ptr Spline::convertSplineToCloud(std::vector<Eigen::Vector3f> &spline)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointXYZ point_temp;
    for (auto const point : spline)
    {
        point_temp.getVector3fMap() = point;
        cloud->push_back(point_temp);
    }
    return cloud;
}

pcl::PointCloud<pcl::PointXYZ>::Ptr Spline::computeApproximatedBSplines(std::vector<Eigen::Vector3f> &vector_path)
{
    std::vector<Eigen::Vector3f> spline;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_output(new pcl::PointCloud<pcl::PointXYZ>);

    spline = SplineApproximation(vector_path);
    cloud_output = convertSplineToCloud(spline);

    return cloud_output;
}

pcl::PointCloud<pcl::PointXYZ>::Ptr Spline::compute(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_points)
{

    std::vector<Eigen::Vector3f> vec;
    for (size_t i = 0; i < cloud_points->size(); ++i)
    {
        vec.push_back(cloud_points->points[i].getVector3fMap());
    }

    pcl::PointCloud<pcl::PointXYZ>::Ptr output = computeApproximatedBSplines(vec);
    return output;
}
